#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `awsme` package."""

from click.testing import CliRunner
from moto import mock_cloudwatch

from awsme.cli import main
from awsme import create_cloud_watch


@mock_cloudwatch
def test_command_line_interface():
    """Test the CLI."""
    runner = CliRunner()
    result = runner.invoke(main)
    assert result.exit_code == 0
    assert 'Successfully sent metric' in result.output
    help_result = runner.invoke(main, ['--help'])
    assert help_result.exit_code == 0
    assert '--help  Show this message and exit.' in help_result.output


@mock_cloudwatch
def test_put_metric():
    cloud_watch = create_cloud_watch(
        'Test Namespace',
        buffered=False,
        dummy=False,
        dimensions={'By intent': 'Test'},
    )
    cloud_watch.log('awsme-test', {'By source': 'awsme'})



import datetime

from awsme.metric import Metric


def test_sample_metric_formatting():
    metric = Metric(
        event_time=datetime.datetime(2014, 4, 14, 14, 4, 14),
        name='app-start',
        dimensions={
            'app': 'makar',
            'config': 'alpha-bch-x8-taker',
        },
    )
    data = metric.to_metric_data()
    assert data == [
        {
            'Dimensions': [
                {'Name': 'app', 'Value': 'makar'},
            ],
            'MetricName': 'app-start',
            'StorageResolution': 60,
            'Timestamp': '2014-04-14 14:04:14.000000 UTC',
            'Unit': 'Count',
            'Value': 1.0
        },
        {
            'Dimensions': [
                {'Name': 'config', 'Value': 'alpha-bch-x8-taker'},
            ],
            'MetricName': 'app-start',
            'StorageResolution': 60,
            'Timestamp': '2014-04-14 14:04:14.000000 UTC',
            'Unit': 'Count',
            'Value': 1.0
        },
        {
            'Dimensions': [
                {'Name': 'app', 'Value': 'makar'},
                {'Name': 'config', 'Value': 'alpha-bch-x8-taker'},
            ],
            'MetricName': 'app-start',
            'StorageResolution': 60,
            'Timestamp': '2014-04-14 14:04:14.000000 UTC',
            'Unit': 'Count',
            'Value': 1.0
        }
    ]



"""Metrics client factory"""

from typing import Dict, Optional

from .cloud_watch import CloudWatch
from .dummy import DummyCloudWatch
from .immediate_recorder import ImmediateRecorder
from .buffered_recorder import BufferedRecorder
from .async_recorder import AsyncRecorder


def create_cloud_watch(namespace: str,
                       asynchronous: Optional[bool] = True,
                       buffered: Optional[bool] = True,
                       dummy: Optional[bool] = False,
                       dimensions: Optional[Dict[str, str]] = None,
                       **kwargs
                       ) -> CloudWatch:
    """Create CloudWatch client using passed arguments"""
    return CloudWatch(
        namespace,
        dimensions or {},
        recorder_factory(asynchronous, buffered, dummy),
        **kwargs
    )


def recorder_factory(asynchronous: Optional[bool] = True,
                     buffered: Optional[bool] = True,
                     dummy: Optional[bool] = False):
    def create_recorder(namespace, client):
        if dummy:
            return DummyCloudWatch()
        recorder = ImmediateRecorder(
            namespace=namespace,
            client=client,
        )
        if buffered:
            recorder = BufferedRecorder(recorder)
        if asynchronous:
            recorder = AsyncRecorder(recorder)
        return recorder
    return create_recorder



"""Console script to test boto configuration."""

import sys
import logging

import click

from .factory import create_cloud_watch


@click.command()
def main():
    """Send test metric to AWS CloudWatch"""
    logging.basicConfig(level=logging.DEBUG)
    cloud_watch = create_cloud_watch(
        'Test Namespace',
        asynchronous=False,
        buffered=False,
        dummy=False,
        dimensions={'By intent': 'Test'},
    )
    cloud_watch.log('awsme-test', {'By source': 'awsme'})
    print('Successfully sent metric "awsme-test" to "Test Namespace"')
    return 0


if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover



from queue import Queue, Full, Empty
from threading import Thread
from typing import Optional

from .metric import Metric


class AsyncRecorder:
    """Offload actual sending to a thread"""

    def __init__(self, recorder) -> None:
        self._recorder = recorder
        self._queue = Queue(maxsize=1000)  # type: Queue
        self._thread = None  # type: Optional[Thread]
        self._start()

    def put_metric(self, metric: Metric) -> None:
        """Enqueue metric for sending in the background."""
        try:
            self._queue.put(metric, block=False)
        except Full:
            # ignore metric if queue is full:
            pass

    def _start(self):
        """Start sending thread"""
        if self._thread is not None:
            raise RuntimeError("Attempted to start AsyncRecorder twice")
        self._thread = Thread(target=self._background_send)
        self._thread.daemon = True
        self._thread.start()

    def _background_send(self):
        """Send metrics through actual recorder."""
        for metric in self._pending_metrics():
            self._recorder.put_metric(metric)

    def _pending_metrics(self):
        """Infinite metrics generator."""
        while True:
            try:
                metric = self._queue.get(block=True)
            except Empty:
                return
            try:
                yield metric
            finally:
                self._queue.task_done()



import datetime
import logging
from unittest import mock

from awsme.immediate_recorder import ImmediateRecorder
from awsme.metric import Metric
from awsme.buffered_recorder import (
    BufferedRecorder,
    PAGE_SIZE,
    MAX_BATCH_SIZE,
    MAX_BUFFER_SIZE,
)


def test_buffered_recorder_buffers_incomplete_page():
    _, fake_recorder = make_recorder(10)
    assert fake_recorder.called == 0


def test_buffered_recorder_flushes_full_page():
    _, fake_recorder = make_recorder(25)
    fake_recorder.put_metric_data.assert_called_once()
    assert len(fake_recorder.put_metric_data.call_args[0][0]) == PAGE_SIZE


def test_buffered_recorder_rate_control(caplog):
    with caplog.at_level(logging.WARNING):
        _, fake_recorder = make_recorder(MAX_BUFFER_SIZE)
    assert len(caplog.records) == 1
    assert fake_recorder.put_metric_data.call_count == MAX_BATCH_SIZE


def test_buffered_recorder_complete_flush_keeps_nothing():
    recorder, fake_recorder = make_recorder(MAX_BUFFER_SIZE)
    recorder.flush()
    assert fake_recorder.put_metric_data.call_count == MAX_BUFFER_SIZE // PAGE_SIZE


def make_recorder(dimensions_count):
    fake_recorder = mock.Mock(spec=ImmediateRecorder)
    recorder = BufferedRecorder(recorder=fake_recorder)
    recorder.put_metric(Metric(
        datetime.datetime.utcnow(),
        'test',
        {str(number): str(number)
         for number in range(dimensions_count)},
    ))
    return recorder, fake_recorder



import datetime
from typing import List, Dict, Any

import attr


@attr.s(auto_attribs=True)
class Metric:
    """Metric data-class"""

    event_time: datetime.datetime
    name: str
    dimensions: Dict[str, str]
    value: float = 1
    unit: str = 'Count'
    storage_resolution: int = 60

    def to_metric_data(self) -> List[Dict[str, Any]]:
        """Return list of metrics in AWS CloudWatch format"""
        data = [
            self._format_single_metric({key: value})
            for key, value in self.dimensions.items()
        ]
        data.append(self._format_single_metric(self.dimensions))
        return data

    def _format_single_metric(self, dimensions: Dict[str, str]) -> Dict[str, Any]:
        timestamp = self.event_time.strftime(
            '%Y-%m-%d %H:%M:%S.%f UTC'
        )
        value = float(self.value)
        cw_dimensions = [
            {'Name': key, 'Value': value}
            for key, value in dimensions.items()
        ]
        return {
            'MetricName': self.name,
            'Dimensions': cw_dimensions,
            'Timestamp': timestamp,
            'Value': value,
            'Unit': self.unit,
            'StorageResolution': self.storage_resolution,
        }



import datetime
import logging
from typing import Optional, Dict, Callable

import boto3

from .metric import Metric


logger = logging.getLogger(__name__)


class CloudWatch:
    """CloudWatch client, provides method log, that records metric
    using recorder_class instance.
    """

    def __init__(self,
                 namespace: str,
                 default_dimensions: Dict[str, str],
                 recorder_class: Callable,
                 **kwargs) -> None:
        client = boto3.client(
            'cloudwatch',
            **kwargs,
        )
        self.default_dimensions = default_dimensions
        self._recorder = recorder_class(namespace, client)

    def log(self,
            name: str,
            dimensions: Optional[Dict[str, str]] = None,
            value: float = 1,
            **kwargs) -> None:
        """Record metric.

        ``dimensions`` are merged with default_dimensions.
        Keyword arguments are passed to Metric constructor.
        """
        complete_dimensions = self.default_dimensions.copy()
        complete_dimensions.update(dimensions or {})
        self._record_metric(Metric(
            event_time=datetime.datetime.utcnow(),
            name=name,
            value=value,
            dimensions=complete_dimensions,
            **kwargs,
        ))

    def _record_metric(self, metric: Metric) -> None:
        logger.debug('log: %r', metric)
        self._recorder.put_metric(metric)



import logging
from typing import List, Dict, Any

from .metric import Metric


logger = logging.getLogger(__name__)


class ImmediateRecorder:
    """Simple metric recorder that sends metric data immediately"""

    def __init__(self, namespace: str, client: Any) -> None:
        self._namespace = namespace
        self._client = client

    def put_metric(self, metric: Metric) -> None:
        """Convert Metric instance to AWS format and send."""
        self.put_metric_data(metric.to_metric_data())

    def put_metric_data(self, metric_data: List[Dict[str, Any]]) -> None:
        """Send metric data to boto3 client."""
        logger.debug('put metric data: %r', metric_data)
        self._client.put_metric_data(
            Namespace=self._namespace,
            MetricData=metric_data,
        )



"""Buffering recorder"""

from itertools import count
import logging
from typing import List, Dict, Any

from .immediate_recorder import ImmediateRecorder
from .metric import Metric


logger = logging.getLogger(__name__)
PAGE_SIZE = 20
MAX_BATCH_SIZE = 5
MAX_BUFFER_SIZE = PAGE_SIZE * MAX_BATCH_SIZE * 10


class BufferedRecorder:
    """Accumulate metrics in buffer and send them in pages"""

    def __init__(self, *args, recorder=None, **kwargs) -> None:
        self._recorder = recorder or ImmediateRecorder(*args, **kwargs)
        self._buffer = []  # type: List[Dict[str, Any]]

    def put_metric(self, metric: Metric) -> None:
        """Add metric_data to buffer. Send full pages"""
        metric_data = metric.to_metric_data()
        space_left = MAX_BUFFER_SIZE - len(self._buffer)
        if len(metric_data) > space_left:
            logger.warning(
                "Dropping %d metrics, that don't fit into buffer",
                space_left - len(metric_data),
            )
            metric_data = metric_data[:space_left]
        self._buffer += metric_data
        self.flush(complete=False)

    def flush(self, complete=True):
        '''Sends as much data as possible to CloudWatch.

        If complete is set to False, this only sends at most MAX_BATCH_SIZE full pages.
        Otherwise it sends all buffered pages, even last incomplete page.
        This way, it minimizes the API usage at the cost of delaying data.
        '''
        batches = (
            count()
            if complete
            else range(MAX_BATCH_SIZE)
        )
        for _ in batches:
            page = self._pop_batch(partial=complete)
            if page:
                self._recorder.put_metric_data(page)
            else:
                break

    def _pop_batch(self, partial=True):
        """Pop first page from buffer.

        If there are no pages available, return empty list.
        if partial is True, return even incomplete page.
        """
        if partial or len(self._buffer) >= PAGE_SIZE:
            page = self._buffer[:PAGE_SIZE]
            self._buffer = self._buffer[PAGE_SIZE:]
            return page
        return []



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

import os
from setuptools import setup, find_packages


with open('README.rst') as readme_file:
    README = readme_file.read()


with open('HISTORY.rst') as history_file:
    HISTORY = history_file.read()


with open(os.path.join('requirements', 'base.in')) as fp:
    REQUIREMENTS = list(fp)


setup(
    author="Peter Demin",
    author_email='peterdemin@gmail.com',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
    ],
    description="AWS CloudWatch client library to send "
                "metrics conviniently and efficiently",
    entry_points={
        'console_scripts': [
            'awsme-test=awsme.cli:main',
        ],
    },
    install_requires=REQUIREMENTS,
    license="MIT license",
    long_description=README + '\n\n' + HISTORY,
    include_package_data=True,
    keywords='awsme',
    name='awsme',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    setup_requires=['pytest-runner'],
    test_suite='tests',
    tests_require=['pytest'],
    url='https://github.com/peterdemin/awsme',
    version='0.2.1',
    zip_safe=False,
)
